@@ -37,7 +37,10 @@
 **
 ****************************************************************************/
 
+#include "qml/qqmlprivate.h"
+#include "qv4engine_p.h"
 #include "qv4executablecompilationunit_p.h"
+#include "qv4stackframe_p.h"
 
 #include <private/qv4engine_p.h>
 #include <private/qv4regexp_p.h>
@@ -53,6 +56,8 @@
 #include <private/qv4compilationunitmapper_p.h>
 #include <private/qml_compile_hash_p.h>
 #include <private/qqmltypewrapper_p.h>
+#include <private/inlinecomponentutils_p.h>
+#include <private/qv4resolvedtypereference_p.h>
 
 #include <QtQml/qqmlfile.h>
 #include <QtQml/qqmlpropertymap.h>
@@ -62,6 +67,7 @@
 #include <QtCore/qfileinfo.h>
 #include <QtCore/qscopeguard.h>
 #include <QtCore/qcryptographichash.h>
+#include <QtCore/QScopedValueRollback>
 
 #if defined(QML_COMPILE_HASH)
 #  ifdef Q_OS_LINUX
@@ -94,11 +100,15 @@ ExecutableCompilationUnit::~ExecutableCompilationUnit()
 
 QString ExecutableCompilationUnit::localCacheFilePath(const QUrl &url)
 {
+    static const QByteArray envCachePath = qgetenv("QML_DISK_CACHE_PATH");
+
     const QString localSourcePath = QQmlFile::urlToLocalFileOrQrc(url);
     const QString cacheFileSuffix = QFileInfo(localSourcePath + QLatin1Char('c')).completeSuffix();
     QCryptographicHash fileNameHash(QCryptographicHash::Sha1);
     fileNameHash.addData(localSourcePath.toUtf8());
-    QString directory = QStandardPaths::writableLocation(QStandardPaths::CacheLocation) + QLatin1String("/qmlcache/");
+    QString directory = envCachePath.isEmpty()
+            ? QStandardPaths::writableLocation(QStandardPaths::CacheLocation) + QLatin1String("/qmlcache/")
+            : QString::fromLocal8Bit(envCachePath) + QLatin1String("/");
     QDir::root().mkpath(directory);
     return directory + QString::fromUtf8(fileNameHash.result().toHex()) + QLatin1Char('.') + cacheFileSuffix;
 }
@@ -121,7 +131,7 @@ static QString toString(QV4::ReturnedValue v)
 static void dumpConstantTable(const StaticValue *constants, uint count)
 {
     QDebug d = qDebug();
-    d.nospace() << right;
+    d.nospace() << Qt::right;
     for (uint i = 0; i < count; ++i) {
         d << qSetFieldWidth(8) << i << qSetFieldWidth(0) << ":    "
           << toString(constants[i].asReturnedValue()).toUtf8().constData() << "\n";
@@ -201,9 +211,24 @@ QV4::Function *ExecutableCompilationUnit::linkToEngine(ExecutionEngine *engine)
     }
 
     runtimeFunctions.resize(data->functionTableSize);
+    const QQmlPrivate::AOTCompiledFunction *aotFunction = aotCompiledFunctions;
+
+    auto advanceAotFunction = [&](int i) -> const QQmlPrivate::AOTCompiledFunction * {
+        if (aotFunction) {
+            if (aotFunction->functionPtr) {
+                if (aotFunction->index == i)
+                    return aotFunction++;
+            } else {
+                aotFunction = nullptr;
+            }
+        }
+        return nullptr;
+    };
+
     for (int i = 0 ;i < runtimeFunctions.size(); ++i) {
         const QV4::CompiledData::Function *compiledFunction = data->functionAt(i);
-        runtimeFunctions[i] = QV4::Function::create(engine, this, compiledFunction);
+        runtimeFunctions[i] = QV4::Function::create(engine, this, compiledFunction,
+                                                    advanceAotFunction(i));
     }
 
     Scope scope(engine);
@@ -283,7 +308,6 @@ void ExecutableCompilationUnit::unlink()
         Q_ASSERT(data && propertyCaches.count() > 0 && propertyCaches.at(/*root object*/0));
         if (qmlEngine)
             qmlEngine->unregisterInternalCompositeType(this);
-        QQmlMetaType::unregisterInternalCompositeType(this);
         isRegisteredWithEngine = false;
     }
 
@@ -383,51 +407,130 @@ IdentifierHash ExecutableCompilationUnit::createNamedObjectsPerComponent(int com
     return *namedObjectsPerComponentCache.insert(componentObjectIndex, namedObjectCache);
 }
 
-void ExecutableCompilationUnit::finalizeCompositeType(QQmlEnginePrivate *qmlEngine)
+void ExecutableCompilationUnit::finalizeCompositeType(QQmlEnginePrivate *qmlEngine, CompositeMetaTypeIds types)
 {
     this->qmlEngine = qmlEngine;
 
     // Add to type registry of composites
     if (propertyCaches.needsVMEMetaObject(/*root object*/0)) {
-        QQmlMetaType::registerInternalCompositeType(this);
+        // typeIds is only valid for types that have references to themselves.
+        if (!types.isValid())
+            types = CompositeMetaTypeIds::fromCompositeName(rootPropertyCache()->className());
+        typeIds = types;
         qmlEngine->registerInternalCompositeType(this);
+
     } else {
         const QV4::CompiledData::Object *obj = objectAt(/*root object*/0);
         auto *typeRef = resolvedTypes.value(obj->inheritedTypeNameIndex);
         Q_ASSERT(typeRef);
-        if (typeRef->compilationUnit) {
-            metaTypeId = typeRef->compilationUnit->metaTypeId;
-            listMetaTypeId = typeRef->compilationUnit->listMetaTypeId;
+        if (const auto compilationUnit = typeRef->compilationUnit()) {
+            typeIds = compilationUnit->typeIds;
         } else {
-            metaTypeId = typeRef->type.typeId();
-            listMetaTypeId = typeRef->type.qListTypeId();
+            const auto type = typeRef->type();
+            typeIds = CompositeMetaTypeIds{ type.typeId(), type.qListTypeId() };
         }
     }
 
     // Collect some data for instantiation later.
+    using namespace  icutils;
+    std::vector<QV4::CompiledData::InlineComponent> allICs {};
+    for (int i=0; i != objectCount(); ++i) {
+        const CompiledObject *obj = objectAt(i);
+        for (auto it = obj->inlineComponentsBegin(); it != obj->inlineComponentsEnd(); ++it) {
+            allICs.push_back(*it);
+        }
+    }
+    std::vector<Node> nodes;
+    nodes.resize(allICs.size());
+    std::iota(nodes.begin(), nodes.end(), 0);
+    AdjacencyList adjacencyList;
+    adjacencyList.resize(nodes.size());
+    fillAdjacencyListForInlineComponents(this, adjacencyList, nodes, allICs);
+    bool hasCycle = false;
+    auto nodesSorted = topoSort(nodes, adjacencyList, hasCycle);
+    Q_ASSERT(!hasCycle); // would have already been discovered by qqmlpropertycachcecreator
+
+    // We need to first iterate over all inline components, as the containing component might create instances of them
+    // and in that case we need to add its object count
+    for (auto nodeIt = nodesSorted.rbegin(); nodeIt != nodesSorted.rend(); ++nodeIt) {
+        const auto &ic = allICs.at(nodeIt->index);
+        int lastICRoot = ic.objectIndex;
+        for (int i = ic.objectIndex; i<objectCount(); ++i) {
+            const QV4::CompiledData::Object *obj = objectAt(i);
+            bool leftCurrentInlineComponent =
+                       (i != lastICRoot && obj->flags & QV4::CompiledData::Object::IsInlineComponentRoot)
+                    || !(obj->flags & QV4::CompiledData::Object::InPartOfInlineComponent);
+            if (leftCurrentInlineComponent)
+                break;
+            inlineComponentData[lastICRoot].totalBindingCount += obj->nBindings;
+
+            if (auto *typeRef = resolvedTypes.value(obj->inheritedTypeNameIndex)) {
+                const auto type = typeRef->type();
+                if (type.isValid() && type.parserStatusCast() != -1)
+                    ++inlineComponentData[lastICRoot].totalParserStatusCount;
+
+                ++inlineComponentData[lastICRoot].totalObjectCount;
+                if (const auto compilationUnit = typeRef->compilationUnit()) {
+                    // if the type is an inline component type, we have to extract the information from it
+                    // This requires that inline components are visited in the correct order
+                    auto icRoot = compilationUnit->icRoot;
+                    if (type.isInlineComponentType())
+                        icRoot = type.inlineComponentId();
+                    QScopedValueRollback<int> rollback {compilationUnit->icRoot, icRoot};
+                    inlineComponentData[lastICRoot].totalBindingCount += compilationUnit->totalBindingsCount();
+                    inlineComponentData[lastICRoot].totalParserStatusCount += compilationUnit->totalParserStatusCount();
+                    inlineComponentData[lastICRoot].totalObjectCount += compilationUnit->totalObjectCount();
+                }
+            }
+        }
+    }
     int bindingCount = 0;
     int parserStatusCount = 0;
     int objectCount = 0;
     for (quint32 i = 0, count = this->objectCount(); i < count; ++i) {
         const QV4::CompiledData::Object *obj = objectAt(i);
+        if (obj->flags & QV4::CompiledData::Object::InPartOfInlineComponent) {
+            continue;
+        }
         bindingCount += obj->nBindings;
         if (auto *typeRef = resolvedTypes.value(obj->inheritedTypeNameIndex)) {
-            if (typeRef->type.isValid()) {
-                if (typeRef->type.parserStatusCast() != -1)
-                    ++parserStatusCount;
-            }
+            const auto type = typeRef->type();
+            if (type.isValid() && type.parserStatusCast() != -1)
+                ++parserStatusCount;
             ++objectCount;
-            if (typeRef->compilationUnit) {
-                bindingCount += typeRef->compilationUnit->totalBindingsCount;
-                parserStatusCount += typeRef->compilationUnit->totalParserStatusCount;
-                objectCount += typeRef->compilationUnit->totalObjectCount;
+            if (const auto compilationUnit = typeRef->compilationUnit()) {
+                auto icRoot = compilationUnit->icRoot;
+                if (type.isInlineComponentType())
+                    icRoot = type.inlineComponentId();
+                QScopedValueRollback<int> rollback {compilationUnit->icRoot, icRoot};
+                bindingCount += compilationUnit->totalBindingsCount();
+                parserStatusCount += compilationUnit->totalParserStatusCount();
+                objectCount += compilationUnit->totalObjectCount();
             }
         }
     }
 
-    totalBindingsCount = bindingCount;
-    totalParserStatusCount = parserStatusCount;
-    totalObjectCount = objectCount;
+    m_totalBindingsCount = bindingCount;
+    m_totalParserStatusCount = parserStatusCount;
+    m_totalObjectCount = objectCount;
+}
+
+int ExecutableCompilationUnit::totalBindingsCount() const {
+    if (icRoot == -1)
+        return m_totalBindingsCount;
+    return inlineComponentData[icRoot].totalBindingCount;
+}
+
+int ExecutableCompilationUnit::totalObjectCount() const {
+    if (icRoot == -1)
+        return m_totalObjectCount;
+    return inlineComponentData[icRoot].totalObjectCount;
+}
+
+int ExecutableCompilationUnit::totalParserStatusCount() const {
+    if (icRoot == -1)
+        return m_totalParserStatusCount;
+    return inlineComponentData[icRoot].totalParserStatusCount;
 }
 
 bool ExecutableCompilationUnit::verifyChecksum(const CompiledData::DependentTypesHasher &dependencyHasher) const
@@ -445,6 +548,13 @@ bool ExecutableCompilationUnit::verifyChecksum(const CompiledData::DependentType
                       sizeof(data->dependencyMD5Checksum)) == 0;
 }
 
+CompositeMetaTypeIds ExecutableCompilationUnit::typeIdsForComponent(int objectid) const
+{
+    if (objectid == 0)
+        return typeIds;
+    return inlineComponentData[objectid].typeIds;
+}
+
 QStringList ExecutableCompilationUnit::moduleRequests() const
 {
     QStringList requests;
@@ -704,68 +814,6 @@ bool ExecutableCompilationUnit::saveToDisk(const QUrl &unitUrl, QString *errorSt
     });
 }
 
-/*!
-Returns the property cache, if one alread exists.  The cache is not referenced.
-*/
-QQmlRefPointer<QQmlPropertyCache> ResolvedTypeReference::propertyCache() const
-{
-    if (type.isValid())
-        return typePropertyCache;
-    else
-        return compilationUnit->rootPropertyCache();
-}
-
-/*!
-Returns the property cache, creating one if it doesn't already exist.  The cache is not referenced.
-*/
-QQmlRefPointer<QQmlPropertyCache> ResolvedTypeReference::createPropertyCache(QQmlEngine *engine)
-{
-    if (typePropertyCache) {
-        return typePropertyCache;
-    } else if (type.isValid()) {
-        typePropertyCache = QQmlEnginePrivate::get(engine)->cache(type.metaObject(), minorVersion);
-        return typePropertyCache;
-    } else {
-        return compilationUnit->rootPropertyCache();
-    }
-}
-
-bool ResolvedTypeReference::addToHash(QCryptographicHash *hash, QQmlEngine *engine)
-{
-    if (type.isValid()) {
-        bool ok = false;
-        hash->addData(createPropertyCache(engine)->checksum(&ok));
-        return ok;
-    }
-    if (!compilationUnit)
-        return false;
-    hash->addData(compilationUnit->data->md5Checksum,
-                  sizeof(compilationUnit->data->md5Checksum));
-    return true;
-}
-
-template <typename T>
-bool qtTypeInherits(const QMetaObject *mo) {
-    while (mo) {
-        if (mo == &T::staticMetaObject)
-            return true;
-        mo = mo->superClass();
-    }
-    return false;
-}
-
-void ResolvedTypeReference::doDynamicTypeCheck()
-{
-    const QMetaObject *mo = nullptr;
-    if (typePropertyCache)
-        mo = typePropertyCache->firstCppMetaObject();
-    else if (type.isValid())
-        mo = type.metaObject();
-    else if (compilationUnit)
-        mo = compilationUnit->rootPropertyCache()->firstCppMetaObject();
-    isFullyDynamicType = qtTypeInherits<QQmlPropertyMap>(mo);
-}
-
 bool ResolvedTypeReferenceMap::addToHash(QCryptographicHash *hash, QQmlEngine *engine) const
 {
     for (auto it = constBegin(), end = constEnd(); it != end; ++it) {
@@ -779,24 +827,8 @@ bool ResolvedTypeReferenceMap::addToHash(QCryptographicHash *hash, QQmlEngine *e
 QString ExecutableCompilationUnit::bindingValueAsString(const CompiledData::Binding *binding) const
 {
     using namespace CompiledData;
+#if QT_CONFIG(translation)
     switch (binding->type) {
-    case Binding::Type_Script:
-    case Binding::Type_String:
-        return stringAt(binding->stringIndex);
-    case Binding::Type_Null:
-        return QStringLiteral("null");
-    case Binding::Type_Boolean:
-        return binding->value.b ? QStringLiteral("true") : QStringLiteral("false");
-    case Binding::Type_Number:
-        return QString::number(bindingValueAsNumber(binding), 'g', QLocale::FloatingPointShortest);
-    case Binding::Type_Invalid:
-        return QString();
-#if !QT_CONFIG(translation)
-    case Binding::Type_TranslationById:
-    case Binding::Type_Translation:
-        return stringAt(
-                data->translations()[binding->value.translationDataIndex].stringIndex);
-#else
     case Binding::Type_TranslationById: {
         const TranslationData &translation
                 = data->translations()[binding->value.translationDataIndex];
@@ -809,27 +841,19 @@ QString ExecutableCompilationUnit::bindingValueAsString(const CompiledData::Bind
         // This code must match that in the qsTr() implementation
         const QString &path = fileName();
         int lastSlash = path.lastIndexOf(QLatin1Char('/'));
-        QStringRef context = (lastSlash > -1) ? path.midRef(lastSlash + 1, path.length() - lastSlash - 5)
-                                              : QStringRef();
+        QStringView context = (lastSlash > -1) ? QStringView{path}.mid(lastSlash + 1, path.length() - lastSlash - 5)
+                                              : QStringView();
         QByteArray contextUtf8 = context.toUtf8();
         QByteArray comment = stringAt(translation.commentIndex).toUtf8();
         QByteArray text = stringAt(translation.stringIndex).toUtf8();
         return QCoreApplication::translate(contextUtf8.constData(), text.constData(),
                                            comment.constData(), translation.number);
     }
-#endif
     default:
         break;
     }
-    return QString();
-}
-
-QString ExecutableCompilationUnit::bindingValueAsScriptString(
-        const CompiledData::Binding *binding) const
-{
-    return (binding->type == CompiledData::Binding::Type_String)
-            ? CompiledData::Binding::escapedString(stringAt(binding->stringIndex))
-            : bindingValueAsString(binding);
+#endif
+    return CompilationUnit::bindingValueAsString(binding);
 }
 
 bool ExecutableCompilationUnit::verifyHeader(
