@@ -32,6 +32,7 @@
 #include <QtQml/qqmlcontext.h>
 #include <QtQml/qqmlcomponent.h>
 #include <private/qmetaobjectbuilder_p.h>
+#include <private/qqmlcontextdata_p.h>
 #include <QCryptographicHash>
 #include "../../shared/util.h"
 
@@ -55,29 +56,109 @@ private slots:
     void metaObjectSize();
     void metaObjectChecksum();
     void metaObjectsForRootElements();
+    void derivedGadgetMethod();
+    void restrictRegistrationVersion();
 
 private:
     QQmlEngine engine;
 };
 
+class BaseGadget
+{
+    Q_GADGET
+    QML_ANONYMOUS
+public:
+    Q_INVOKABLE QString stringValue() { return QLatin1String("base"); }
+};
+
+class DerivedGadget : public BaseGadget
+{
+    Q_GADGET
+    QML_ANONYMOUS
+public:
+    Q_INVOKABLE QString stringValue() { return QLatin1String("derived"); }
+};
+
+class GadgetUser : public QObject
+{
+    Q_OBJECT
+    Q_PROPERTY(BaseGadget base READ base CONSTANT)
+    Q_PROPERTY(DerivedGadget derived READ derived CONSTANT)
+    Q_PROPERTY(QString baseString READ baseString WRITE setBaseString NOTIFY baseStringChanged)
+    Q_PROPERTY(QString derivedString READ derivedString WRITE setDerivedString NOTIFY derivedStringChanged)
+    QML_ELEMENT
+
+public:
+    BaseGadget base() const { return m_base; }
+    DerivedGadget derived() const { return m_derived; }
+    QString baseString() const { return m_baseString; }
+    QString derivedString() const { return m_derivedString; }
+
+public slots:
+    void setBaseString(QString baseString)
+    {
+        if (m_baseString == baseString)
+            return;
+
+        m_baseString = baseString;
+        emit baseStringChanged(m_baseString);
+    }
+
+    void setDerivedString(QString derivedString)
+    {
+        if (m_derivedString == derivedString)
+            return;
+
+        m_derivedString = derivedString;
+        emit derivedStringChanged(m_derivedString);
+    }
+
+signals:
+    void baseStringChanged(QString baseString);
+    void derivedStringChanged(QString derivedString);
+
+private:
+    BaseGadget m_base;
+    DerivedGadget m_derived;
+    QString m_baseString;
+    QString m_derivedString;
+};
+
 class BaseObject : public QObject
 {
     Q_OBJECT
+    QML_ELEMENT
     Q_PROPERTY(int propertyA READ propertyA NOTIFY propertyAChanged)
     Q_PROPERTY(QString propertyB READ propertyB NOTIFY propertyBChanged)
+    Q_PROPERTY(int highVersion READ highVersion WRITE setHighVersion NOTIFY highVersionChanged REVISION(4, 0))
+
 public:
     BaseObject(QObject *parent = nullptr) : QObject(parent) {}
 
     int propertyA() const { return 0; }
     QString propertyB() const { return QString(); }
+    int highVersion() const { return m_highVersion; }
 
 public Q_SLOTS:
     void slotA() {}
 
+    void setHighVersion(int highVersion)
+    {
+        if (m_highVersion == highVersion)
+            return;
+
+        m_highVersion = highVersion;
+        emit highVersionChanged();
+    }
+
 Q_SIGNALS:
     void propertyAChanged();
     void propertyBChanged();
     void signalA();
+    void highVersionChanged();
+
+private:
+    int m_highVersion = 0;
 };
 
 class DerivedObject : public BaseObject
@@ -114,7 +195,8 @@ void tst_qqmlpropertycache::properties()
     DerivedObject object;
     const QMetaObject *metaObject = object.metaObject();
 
-    QQmlRefPointer<QQmlPropertyCache> cache(new QQmlPropertyCache(metaObject));
+    QQmlRefPointer<QQmlPropertyCache> cache(new QQmlPropertyCache(metaObject),
+                                            QQmlRefPointer<QQmlPropertyCache>::Adopt);
     QQmlPropertyData *data;
 
     QVERIFY((data = cacheProperty(cache, "propertyA")));
@@ -136,8 +218,11 @@ void tst_qqmlpropertycache::propertiesDerived()
     DerivedObject object;
     const QMetaObject *metaObject = object.metaObject();
 
-    QQmlRefPointer<QQmlPropertyCache> parentCache(new QQmlPropertyCache(&BaseObject::staticMetaObject));
-    QQmlRefPointer<QQmlPropertyCache> cache(parentCache->copyAndAppend(object.metaObject()));
+    QQmlRefPointer<QQmlPropertyCache> parentCache(
+                new QQmlPropertyCache(&BaseObject::staticMetaObject),
+                QQmlRefPointer<QQmlPropertyCache>::Adopt);
+    QQmlRefPointer<QQmlPropertyCache> cache(parentCache->copyAndAppend(object.metaObject()),
+                                            QQmlRefPointer<QQmlPropertyCache>::Adopt);
     QQmlPropertyData *data;
 
     QVERIFY((data = cacheProperty(cache, "propertyA")));
@@ -161,8 +246,11 @@ void tst_qqmlpropertycache::revisionedProperties()
     DerivedObject object;
     const QMetaObject *metaObject = object.metaObject();
 
-    QQmlRefPointer<QQmlPropertyCache> cacheWithoutVersion(new QQmlPropertyCache(metaObject));
-    QQmlRefPointer<QQmlPropertyCache> cacheWithVersion(new QQmlPropertyCache(metaObject, 1));
+    QQmlRefPointer<QQmlPropertyCache> cacheWithoutVersion(new QQmlPropertyCache(metaObject),
+                                                          QQmlRefPointer<QQmlPropertyCache>::Adopt);
+    QQmlRefPointer<QQmlPropertyCache> cacheWithVersion(
+                new QQmlPropertyCache(metaObject, QTypeRevision::fromMinorVersion(1)),
+                QQmlRefPointer<QQmlPropertyCache>::Adopt);
     QQmlPropertyData *data;
 
     QVERIFY((data = cacheProperty(cacheWithoutVersion, "propertyE")));
@@ -176,7 +264,8 @@ void tst_qqmlpropertycache::methods()
     DerivedObject object;
     const QMetaObject *metaObject = object.metaObject();
 
-    QQmlRefPointer<QQmlPropertyCache> cache(new QQmlPropertyCache(metaObject));
+    QQmlRefPointer<QQmlPropertyCache> cache(new QQmlPropertyCache(metaObject),
+                                            QQmlRefPointer<QQmlPropertyCache>::Adopt);
     QQmlPropertyData *data;
 
     QVERIFY((data = cacheProperty(cache, "slotA")));
@@ -210,8 +299,11 @@ void tst_qqmlpropertycache::methodsDerived()
     DerivedObject object;
     const QMetaObject *metaObject = object.metaObject();
 
-    QQmlRefPointer<QQmlPropertyCache> parentCache(new QQmlPropertyCache(&BaseObject::staticMetaObject));
-    QQmlRefPointer<QQmlPropertyCache> cache(parentCache->copyAndAppend(object.metaObject()));
+    QQmlRefPointer<QQmlPropertyCache> parentCache(
+                new QQmlPropertyCache(&BaseObject::staticMetaObject),
+                QQmlRefPointer<QQmlPropertyCache>::Adopt);
+    QQmlRefPointer<QQmlPropertyCache> cache(parentCache->copyAndAppend(object.metaObject()),
+                                            QQmlRefPointer<QQmlPropertyCache>::Adopt);
     QQmlPropertyData *data;
 
     QVERIFY((data = cacheProperty(cache, "slotA")));
@@ -245,7 +337,8 @@ void tst_qqmlpropertycache::signalHandlers()
     DerivedObject object;
     const QMetaObject *metaObject = object.metaObject();
 
-    QQmlRefPointer<QQmlPropertyCache> cache(new QQmlPropertyCache(metaObject));
+    QQmlRefPointer<QQmlPropertyCache> cache(new QQmlPropertyCache(metaObject),
+                                            QQmlRefPointer<QQmlPropertyCache>::Adopt);
     QQmlPropertyData *data;
 
     QVERIFY((data = cacheProperty(cache, "onSignalA")));
@@ -273,8 +366,11 @@ void tst_qqmlpropertycache::signalHandlersDerived()
     DerivedObject object;
     const QMetaObject *metaObject = object.metaObject();
 
-    QQmlRefPointer<QQmlPropertyCache> parentCache(new QQmlPropertyCache(&BaseObject::staticMetaObject));
-    QQmlRefPointer<QQmlPropertyCache> cache(parentCache->copyAndAppend(object.metaObject()));
+    QQmlRefPointer<QQmlPropertyCache> parentCache(
+                new QQmlPropertyCache(&BaseObject::staticMetaObject),
+                QQmlRefPointer<QQmlPropertyCache>::Adopt);
+    QQmlRefPointer<QQmlPropertyCache> cache(parentCache->copyAndAppend(object.metaObject()),
+                                            QQmlRefPointer<QQmlPropertyCache>::Adopt);
     QQmlPropertyData *data;
 
     QVERIFY((data = cacheProperty(cache, "onSignalA")));
@@ -418,7 +514,7 @@ void tst_qqmlpropertycache::passQGadget()
     QVERIFY(before.isNull());
     emit emitter.emitGadget(SimpleGadget());
     QVariant after = obj->property("result");
-    QCOMPARE(QMetaType::Type(after.type()), QMetaType::Bool);
+    QCOMPARE(after.typeId(), QMetaType::Bool);
     QVERIFY(after.toBool());
 }
 
@@ -478,7 +574,9 @@ class TestClassWithClassInfo : public QObject
     int(sizeof(arr) / sizeof(arr[0]))
 
 #define TEST_CLASS(Class) \
-    QTest::newRow(#Class) << &Class::staticMetaObject << ARRAY_SIZE(qt_meta_data_##Class) << ARRAY_SIZE(qt_meta_stringdata_##Class.data)
+    QTest::newRow(#Class) \
+            << &Class::staticMetaObject << ARRAY_SIZE(qt_meta_data_##Class) \
+            << int(sizeof(qt_meta_stringdata_##Class.offsetsAndSize) / (sizeof(uint) * 2))
 
 Q_DECLARE_METATYPE(const QMetaObject*);
 
@@ -554,4 +652,26 @@ void tst_qqmlpropertycache::metaObjectsForRootElements()
     QCOMPARE(obj->property("result").toString(), QString::fromLatin1("good"));
 }
 
+void tst_qqmlpropertycache::derivedGadgetMethod()
+{
+    qmlRegisterTypesAndRevisions<BaseGadget, DerivedGadget, GadgetUser>("Test.PropertyCache", 1);
+    QQmlEngine engine;
+    QQmlComponent c(&engine, testFileUrl("derivedGadgetMethod.qml"));
+    QVERIFY(c.isReady());
+    QScopedPointer<QObject> obj(c.create());
+    QVERIFY(!obj.isNull());
+    GadgetUser *gadgetUser = qobject_cast<GadgetUser *>(obj.data());
+    QVERIFY(gadgetUser);
+    QCOMPARE(gadgetUser->baseString(), QString::fromLatin1("base"));
+    QCOMPARE(gadgetUser->derivedString(), QString::fromLatin1("derived"));
+}
+
+void tst_qqmlpropertycache::restrictRegistrationVersion()
+{
+    qmlRegisterTypesAndRevisions<BaseObject>("Test.PropertyCache", 3);
+    QQmlEngine engine;
+    QQmlComponent c(&engine, testFileUrl("highVersion.qml"));
+    QVERIFY(c.isError());
+}
+
 QTEST_MAIN(tst_qqmlpropertycache)
